class FLATLAND_RL:
    PING = "FLATLAND_RL.PING"
    PONG = "FLATLAND_RL.PONG"

    ENV_CREATE = "FLATLAND_RL.ENV_CREATE"
    ENV_CREATE_RESPONSE = "FLATLAND_RL.ENV_CREATE_RESPONSE"

    ENV_RESET = "FLATLAND_RL.ENV_RESET"
    ENV_RESET_RESPONSE = "FLATLAND_RL.ENV_RESET_RESPONSE"

    ENV_STEP = "FLATLAND_RL.ENV_STEP"
    ENV_STEP_RESPONSE = "FLATLAND_RL.ENV_STEP_RESPONSE"

    ENV_SUBMIT = "FLATLAND_RL.ENV_SUBMIT"
    ENV_SUBMIT_RESPONSE = "FLATLAND_RL.ENV_SUBMIT_RESPONSE"

    ERROR = "FLATLAND_RL.ERROR"
    